/* client.cc, a simple client */

/*  This code is partially based on an example in:  
    _Adventures is UNIX network applications programming_, by
    Rieken and Weiman
*/


/*  This is a very simple client program.  It takes a hostname and a port
    from the command line and attemps a connection.  It is non blocking on
    the socket IO, but blocks on getting input from standard input.  It sends
    the results of the read() from the socket to standard output.
*/

/*  The only way to terminate the process is with CTRL-C or any other
    terminating signal.  It should also gracefully terminate if the server
    closes connection.
*/

#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <limits.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <netinet/in.h>
#include <unistd.h>
#include <signal.h>
#include <limits.h>
#include <iostream.h>
#include <fstream.h>
#include <strings.h>
#include "string2.h"
#include <netdb.h>


int main(int argc, char** argv) {
  int sd;
  struct sockaddr_in name;
  struct hostent *hp;

  if (argc != 3) {
    printf("Usage:  client host_name port_num\n");
    exit(0);
  }//if

  name.sin_family = AF_INET;
  name.sin_port = htons(atoi(argv[2]));

  if (hp = gethostbyname(argv[1])) {
    memcpy(&name.sin_addr.s_addr, hp->h_addr, hp->h_length);
  }//if
  else {
    perror("gethostbyname");
    exit(1);
  }//else

  if ((sd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    perror("socket:");
    exit(1);
  }//if

  if (connect(sd, (struct sockaddr*)(&name), sizeof(name)) < 0) {
    perror("connect");
    exit(1);
  }//if

  int  flags = fcntl(sd, F_GETFL);
  flags |= O_NONBLOCK;
  if (fcntl(sd, F_SETFL, flags) < 0) {
    perror("Fatal error executing nonblock, client");
    exit(1);
  }

  cout << "Connected...\n";
  /* should be good to go now */

  String input(1024);
  String output(1024);

  struct timeval ts; //for sub-second sleeps

  while (TRUE) {
    input.Vclear();
    output.Vclear();
    input.Getline(cin, 1023); //blocks

    //    cout << "Got input: " << input << endl;

    input += "\n";
    if (input.Write(sd, input.Strlen()) < 0) {
      cout << "Write to socket failed.\n";
      break;
    }//if

    //    cout << "Input after Write:  " << input << endl;

    //    cout << "Output before read:  " << output << endl;

    ts.tv_usec = 500000;
    ts.tv_sec = 0;

    select(0, NULL, NULL, NULL, &ts);

/* does NOT block */
    if (output.Read(sd, 1023) < 0) {
      cout << "Read from socket failed.";
      break;
    }//if

    //    cout << "Output after read: " << output << endl;

    cout << output;
  }//while true

  close(sd);

  cout << "Thanks for playing..\n";
}//client










