///**********************  server.cc  ***************************///

/*  This code is based off of a game (ScryMUD, see my web page) that I
    have been writing.  I consider this code public domain, so use it if
    it seems useful.  However, kindly leave my name in the code, along
    with my email and www address.

    EMAIL:  greear@cs.uga.edu
    WWW:    http://www.cs.uga.edu/~greear

*/

/*  This is a simple server.  It handles multiple connections, blocking when
    there is nothing to do (via the select() call).  To increase the
    usefullness of this server, the client_struct should be expanded, and
    the process_input should be made more interesting.  Also, if you don't
    want to block when no input was given, the select() function should NOT
    be given a NULL argument, as it is now.  Instead, this should be a short
    time out (preferable 0).  Then, to keep from busy-waiting, call select
    again, like:

        select(max_desc, NULL, NULL, NULL, sleep_fer_this_time);
    See man pages on select for more info.

    The makefile included here will build a binary called:  bend,
    for ben's daemon.

    NOTE:  I'm adding a fork() call to make this run in the background as
    a daemon because that is what this assignment calls for, but in practice
    this is not needed, just start the process in the background.
    All the stuff with fork and setsid() will be included ONLY when 
    DAEMON_MODE is #defined.  This #define is found in server.h
    To not include that code in the compilation, simply delete the line.

    NOTE2:  The port that this daemon will be listening too will be 4554 by
    default.  If you wish to change it...read a little farther down...

*/



#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <limits.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <netinet/in.h>
#include <unistd.h>
#include <signal.h>
#include <limits.h>
#include <iostream.h>
#include <fstream.h>
#include <strings.h>
#include "string2.h"
#include "list2.h"
#include "client_struct.h"
#include <netdb.h>

#define GRRMUD
#define MAX_HOSTNAME    256
#define MODE_LOG_ME_OFF_PLEASE 100
#define MODE_JUST_FINE_THANKS 101

/* local globals */
int     MAX_DESCRIPTORS_AVAILABLE = 64;  
int     mother_desc = 0;            /* file desc of the mother connection */
int     maxdesc = 0;                    /* highest desc num used */
int     avail_descs = 64;                /* max descriptors available */
int     sockets_connected = 0;
int     do_shutdown = FALSE; 


List<client_struct*>        cs_list;  


int DFLT_PORT = 4554;


	       /* end of global variables */

int get_new_maxdesc(int s);
void run_the_server(int port);
int init_socket(int port);
void server_loop(int s);
void close_sockets(int mother);
int close_socket(client_struct& cs);
int new_descriptor(int s);
int log_off(client_struct& cs);
int get_input(client_struct& cs);
int write_output(client_struct& cs);
void process_input(client_struct& cs, String& input);
void nonblock(int s);
int isvowel(char chr);

/* *********************************************************************
*  main loop and related stuff                                    *
********************************************************************* */

/*
ofstream lg("./Daemon.LOG");
void log(const char* msg) {
  if (!lg) {
    cerr << "ERROR:  lg file not opened!";
  }//if
  else {
    lg << msg << endl << flush;
  }//else
}//log
*/

int isvowel(char chr) {
  return ((chr == 'a') || (chr == 'e') || (chr == 'i') || (chr == 'o') ||
	  (chr == 'u'));
}//isvowel


void process_input(client_struct& cs, String& input) {
  String vowels(100);
  String consonants(100);
  String digits(100);
  String non_alphanum(100);
  String buf(100);

  int len = input.Strlen();
  Sprintf(buf, "In process_Input, len: %i.\n", len);
  log(buf);
  log(input);

  char chr;
  for (int i = 0; i<len; i++) {
    chr = input[i];
    if (isalnum(chr)) {
      if (isdigit(chr)) {
	digits.Append(chr);
      }//if
      else if (isvowel(chr)) {
	vowels.Append(chr);
      }//if
      else {
	consonants.Append(chr);
      }//else
    }//if alpha numeric
    else {
      non_alphanum.Append(chr);
    }//else
  }//for

  input.Vclear();

  log(consonants);


  cs.output += consonants;
  cs.output += digits;
  cs.output += non_alphanum;
  cs.output += vowels;
  cs.output += "\n\nD'ya Grok?>";

  log("Output at end of process input:");
  log(cs.output);
}//process_input


//hopefully trap the TERM signal and exit gracefully
void sig_term_handler(int junk) { 
  log("Sig_term_handler called, shutting down.\n");
  do_shutdown = 1;
}//

int main() {

#ifdef DAEMON_MODE
   pid_t pid;
   if ((pid = fork()) < 0) {
     perror("fork");
     exit(2);
   }//if
   else if (pid != 0) {
     exit(0); //kill parent process
   }//else

   setsid();
   chdir("/"); //cd to root, avoid mount/umount problems

   umask(0); //Clear
#endif

   signal(SIGTERM, (&sig_term_handler));
   String buf(50);

   run_the_server(DFLT_PORT);
   exit(0);
}


/* Init sockets, run game, and cleanup sockets */

void run_the_server(int port) {
   int  s;

   log("Opening mother connection.\n");
  
   mother_desc = s = init_socket(port);
   
   log("Entering game loop.\n");

   server_loop(s);

   close_sockets(s);
   log("Normal termination of the server.\n");
}//run_the_game



/* Does all work, accepts, dis-connects, parse input.. */

void server_loop(int s)  {
   fd_set input_set, output_set, exc_set;
   struct timeval null_time;
   Cell<client_struct*> cs_cell;
   client_struct* cs_ptr;
   client_struct* tmp_ptr;
   int  sockets_connected, i;

   String buf(100);


   /* initialize the time structures */

   null_time.tv_sec = 0;
   null_time.tv_usec = 0;

   maxdesc = s;
   
   avail_descs = MAX_DESCRIPTORS_AVAILABLE;

   log("Entering while loop.\n");

   /* Main loop */
   while (!do_shutdown) {

      FD_ZERO(&input_set);
      FD_ZERO(&output_set);
      FD_ZERO(&exc_set);
      FD_SET(s, &input_set);

      cs_cell = Head(cs_list);
      while (cs_ptr = Next(cs_cell)) {
	 FD_SET(cs_ptr->descriptor, &input_set);
	 FD_SET(cs_ptr->descriptor, &exc_set);
	 FD_SET(cs_ptr->descriptor, &output_set);
      }//while



		       /* do log_offs of all types */

      cs_cell = Head(cs_list);
      cs_ptr = Next(cs_cell);
      while (cs_ptr) {
	log("Doing log-offs.\n");
	if (cs_ptr->mode == MODE_LOG_ME_OFF_PLEASE)  {
	  tmp_ptr = cs_ptr;
	  cs_ptr = Lose(cs_cell);
	  close_socket(*tmp_ptr);
	  maxdesc = get_new_maxdesc(s);
	}//if
	else {
	  cs_ptr = Next(cs_cell);
	}//else
      }//while pc_ptr
      

      
      null_time.tv_sec = 0;
      null_time.tv_usec = 0;

      log("Calling select, poll\n");
      if (select(maxdesc + 1, &input_set, NULL /*&output_set*/, 
		 &exc_set, NULL)
	   < 0) {
         if (errno == EBADF) {
            log("ERROR:  bad file descriptor given in a set to select.\n");
	    continue; //this might fix some strange forced log-off bug
         }//if
         else if (errno == EINTR) {
            log("ERROR:  A non blocked signal was caught.\n");
         }//if
         else {
   	    perror("Select poll");
            log("ERROR: Select poll, n is negative or out of memory.\n");
            exit(1);
         }//else
      }//if

      /* New connection? */
      if (FD_ISSET(s, &input_set)) {
	log("Checking new connections.\n");
	if (new_descriptor(s) < 0) { //set MODE and descriptor
	  perror("New connection");
	}//if
	log("new_descriptor seemed to work.");
      }//if



		       /* kick out link freaks */
      cs_cell = Head(cs_list);
      while (cs_ptr = Next(cs_cell)) {
	log("Losing link freaks.\n");
	if (FD_ISSET(cs_ptr->descriptor, &exc_set)) {
	  FD_CLR(cs_ptr->descriptor, &input_set);
	  FD_CLR(cs_ptr->descriptor, &output_set);
	  log_off(*cs_ptr);
	}//if
      }//while


      log("Get input.\n");
			/* get input */
      cs_cell = Head(cs_list);
      while (cs_ptr = Next(cs_cell)) {
	if (cs_ptr->mode != MODE_LOG_ME_OFF_PLEASE) {
	  if (FD_ISSET(cs_ptr->descriptor, &input_set)) {
	    if (get_input(*cs_ptr) < 0) { 
	      log("Kicked out by negative get_input.\n");
	      log_off(*cs_ptr);
	    }//if
	  }//if
	}//if
      }//while

      int len;
			/* process input */
      cs_cell = Head(cs_list);
      while (cs_ptr = Next(cs_cell)) {
	if (cs_ptr->mode != MODE_LOG_ME_OFF_PLEASE) {
	  len = cs_ptr->input.Strlen();
	  if (len && cs_ptr->input[len - 1] == '\n') {
	    log("About to process input.\n");
	    process_input(*cs_ptr, cs_ptr->input);
	  }//else
	}//if
      }//while

				    /* output */
      cs_cell = Head(cs_list);
      while (cs_ptr = Next(cs_cell)) {
	if (cs_ptr->mode != MODE_LOG_ME_OFF_PLEASE) {
	  log("Doing output.\n");
	  if (write_output(*cs_ptr) < 0) { 
	    log("Kicked out by negative write_output.\n");
	    log_off(*cs_ptr);
	  }//if
	}//if
      }//while
   }//while
}// game_loop


int get_new_maxdesc(int sss) {
   Cell<client_struct*> cell(cs_list);
   client_struct* ptr;
   int retval = sss;

   while (ptr = Next(cell)) {
     if (ptr->descriptor > retval) 
       retval = ptr->descriptor;
   }//while
   return retval;
}//get_new_maxdesc


/* ******************************************************************
*  socket handling                                                  *
****************************************************************** */

int init_socket(int port)   {
   int  s;
   char *opt;
   char hostname[MAX_HOSTNAME+1];
   struct sockaddr_in sa;
   struct hostent *hp;

   log("in init_socket\n");
   bzero(&sa, sizeof(struct sockaddr_in ));

   if (gethostname(hostname, MAX_HOSTNAME) < 0) {
     perror("gethostname");
     exit(1);
   }//if

   hp = gethostbyname(hostname);
   if (hp == NULL) {
      perror("gethostbyname");
      do_shutdown = TRUE;
      exit(1);
   }//if
   sa.sin_family = hp->h_addrtype;
   sa.sin_port   = htons(port);

   if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
      perror("Init-socket");
      exit(1);
   }//if

   if (setsockopt (s, SOL_SOCKET, SO_REUSEADDR, (char *)(&opt), sizeof(opt))
	  < 0) {
      perror ("setsockopt");
      exit (1);
   }//if

   if (bind(s, (struct sockaddr *) & sa, sizeof(sa)) < 0) {
      perror("bind");
      close(s);
      exit(1);
   }//if
   listen(s, 5);
   return(s);
}//init_socket, for the mother descriptor



int new_connection(int s)  {
   struct sockaddr_in isa;
   int  i, t;
 
   log("in new_connection\n");

   i = sizeof(isa);
   if (getsockname(s, (struct sockaddr *) & isa, &i) < 0) {
      perror("getsockname, in new_connection");
      return (-1);
   }//if

   if ((t = accept(s, (struct sockaddr *)(&isa), &i)) < 0) {
      perror("Accept");
      return(-1);
   }//if
   nonblock(t);
   return(t);
}//new_connection()



int new_descriptor(int s) {
   int  desc;
   client_struct *newd = new client_struct;
   int size, i;
   struct sockaddr_in sock;
   struct hostent *from;
   String buf2(100);

   log("In new_descriptor\n");

   if ((desc = new_connection(s)) < 0) {
      delete newd;
      return -1;
   }//if

   newd->descriptor = desc;
   
   if (sockets_connected > avail_descs) {
      log("ERROR:  server was full.\n");
      close(desc);
      delete newd;
      return -1;
   }//if

   if (desc > maxdesc) {
      maxdesc = desc;
   }//if

   log("Finding info.\n");

   /* find info */
   size = sizeof(sock);
   if (getpeername(desc, (struct sockaddr *) & sock, &size) < 0) {
      perror("getpeername");
   }// if 
   else {
      if (!(from = gethostbyaddr((char *)&sock.sin_addr,
		  sizeof(sock.sin_addr), AF_INET))) { //begin of if
	 perror("gethostbyaddr");
	 i = sock.sin_addr.s_addr;
	 Sprintf(newd->host, "%i.%i.%i.%i", (i & 0x000000FF),
	      ((i & 0x0000FF00) >> 8), ((i & 0x00FF0000) >> 16), 
	      ((i & 0xFF000000) >> 24));
      } //if
      else {
	 newd->host = from->h_name;
      }//else
   }//else

   Push(newd, cs_list);
   sockets_connected++;

   return TRUE;
}// new_descriptor()



int write_output(client_struct& cs) {
   int ret = 0;
   int len = cs.output.Strlen();

   String buf(100);
   Sprintf(buf, "Descriptor:  %i:\n%S:EOS\n", cs.descriptor, &(cs.output));
   log(buf);

   if (len) { //no need to write zip
      ret = cs.output.Write(cs.descriptor, len);
   }//if

   Sprintf(buf, "Wrote from server:  %i bytes.\n", ret);
   log(buf);

   return ret;
}//write_output


void    nonblock(int s) {
   int  flags;

   log("in nonblock\n");

   flags = fcntl(s, F_GETFL);
   flags |= O_NONBLOCK;
   if (fcntl(s, F_SETFL, flags) < 0) {
      perror("Fatal error executing nonblock");
      exit(1);
   }
}


int get_input(client_struct& cs)  {
   int ret;
   ret = cs.input.Read(cs.descriptor, MAX_INPUT_LEN);

   log("End of get_input, here it is:");
   log(cs.input);

   return ret;  //valid input added to input buffer, can go parse it now.
}//get_input


void close_sockets(int mother) {
   Cell<client_struct*> cell(cs_list);
   client_struct* ptr;

   log("Closing all sockets now.\n");

   while (ptr = Next(cell)) {
     if (close_socket(*ptr) < 0)
       log("ERROR, close socket failed.\n");
   }//while
   close(mother);
}//close_sockets



int close_socket(client_struct& cs) {
   String buf(100);
   
   log("in close_socket\n");

   if (cs.descriptor != -1) {
     write_output(cs); //flush last little bit of data
     if (close(cs.descriptor) < 0) {
       log("ERROR:  Bad close in close_socket.\n");
     }//if
     else
       sockets_connected--;
   }//if

   delete &cs;
}//close_socket   


int log_off(client_struct& cs) {
   
  //log("in log_off\n");

   cs.mode = MODE_LOG_ME_OFF_PLEASE; //next while loop will do em.
   return 1;
}//log_off






